
{*******************************************************}
{                                                       }
{       Demo: RTF to PDF conversion.                    }
{                                                       }
{       RVReportHelper (http://www.trichview.com)       }
{       loads RTF file and renders the document.        }
{       LLPDFLib (http://www.llion.net)                 }
{       converts RVReportHelper's output to PDF file.   }
{                                                       }
{*******************************************************}

{ Version 2.3

  New in v2.3
  - Compatibility with TRichView 13.4
  New in v2.2
  - LLPDFLib 4 is used instead of LLPDFLib 3
  New in v2.0
  - header and footer
  - hyperlinks (URL and local links)
  - reading margins, header and footer positions, orientation from RTF
}

unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ShellApi, ExtCtrls,
  RVScroll, RichView, PtblRV, RVReport, RVStyle, CRVData, RVTypes,
  llPDFDocument, llPDFTypes, llPDFCanvas;

type

  TPageInfo = record
    PageSize: TSize;
    Doc, Header, Footer: TRect;
  end;

  TCheckpointLocation = class (TCollectionItem)
    public
      Name: String;
      PageIndex, Top: Integer;
  end;
  TLinkToCheckpoint = class (TCollectionItem)
    public
      CheckpointName: String;
      PageIndex: Integer;
      Area: TRect;
  end;

  TForm1 = class(TForm)
    pdf: TPDFDocument;
    btnSave: TButton;
    rvs: TRVStyle;
    rvrh: TRVReportHelper;
    btnOpen: TButton;
    od: TOpenDialog;
    sd: TSaveDialog;
    GroupBox1: TGroupBox;
    Label1: TLabel;
    cmbPageSize: TComboBox;
    rgPageOrientation: TRadioGroup;
    lblStatus: TLabel;
    Image1: TImage;
    Image2: TImage;
    GroupBox2: TGroupBox;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    txtLeft: TEdit;
    txtTop: TEdit;
    txtRight: TEdit;
    txtBottom: TEdit;
    GroupBox3: TGroupBox;
    Label6: TLabel;
    Label7: TLabel;
    txtHeaderY: TEdit;
    txtFooterY: TEdit;
    rvrhHeader: TRVReportHelper;
    rvrhFooter: TRVReportHelper;
    procedure btnSaveClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure btnOpenClick(Sender: TObject);
    procedure Image1Click(Sender: TObject);
    procedure Image2Click(Sender: TObject);
    procedure rvrhDrawHyperlink(Sender: TRVReportHelper;
      RVData: TCustomRVData; ItemNo: Integer; R: TRect);
    procedure rvrhHeaderDrawHyperlink(Sender: TRVReportHelper;
      RVData: TCustomRVData; ItemNo: Integer; R: TRect);
    procedure rvrhFooterDrawHyperlink(Sender: TRVReportHelper;
      RVData: TCustomRVData; ItemNo: Integer; R: TRect);
    procedure rvrhDrawCheckpoint(Sender: TRVReportHelper;
      RVData: TCustomRVData; ItemNo, X, Y: Integer);
  private
    { Private declarations }
    FCurPage: TPDFPage;
    FCurPageIndex: Integer;
    FCheckpoints: TCollection;
    FLinksToCheckpoints: TCollection;
    FPageInfo: TPageInfo;
    procedure rvrhRichViewReadHyperlink(Sender: TCustomRichView;
      const Target, Extras: String; DocFormat: TRVLoadFormat; var StyleNo: Integer;
      var ItemTag: TRVTag; var ItemName: TRVRawByteString);
    function GetDistance(edit: TEdit; ppi: Integer): Integer;
    function GetPageSize: TPDFPageSize;
    function GetPageOrientation: TPDFPageOrientation;
    procedure ShowText(s: String; RedrawForm: Boolean);
    procedure ShowError(s: String);
    procedure CalcPageLayout(Canvas: TCanvas; var PageInfo: TPageInfo);
    procedure InitHelper(h: TRVReportHelper);
    procedure DrawPage(Canvas: TCanvas; PageNo: Integer;
      const PageInfo: TPageInfo);
    procedure SetPDFURL(RVData: TCustomRVData; ItemNo: Integer; R, DocRect: TRect);
    function FindCheckpoint(const CheckpointName: String;
      var PageIndex, Top: Integer): Boolean;
    procedure SaveLinksToCheckpoints;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}
{ Assigning default values to ReportHelper }
procedure TForm1.InitHelper(h: TRVReportHelper);
begin
  h.RichView.Style := rvs;
  h.RichView.RTFReadProperties.TextStyleMode := rvrsAddIfNeeded;
  h.RichView.RTFReadProperties.ParaStyleMode := rvrsAddIfNeeded;
  h.RichView.RTFReadProperties.UnicodeMode := rvruOnlyUnicode;
  h.MetafileCompatibility := True;
  h.RichView.OnReadHyperlink := rvrhRichViewReadHyperlink;
end;
{------------------------------------------------------------------------------}
{ Initialization }
procedure TForm1.FormCreate(Sender: TObject);
begin
  InitHelper(rvrh);
  InitHelper(rvrhHeader);
  InitHelper(rvrhFooter);
  rvrh.RichView.RTFReadProperties.SetHeader(rvrhHeader.RichView.RVData);
  rvrh.RichView.RTFReadProperties.SetFooter(rvrhFooter.RichView.RVData);
  rvrh.RichView.RTFReadProperties.ReadDocParameters := True;
  rvrh.RichView.DocParameters.Units := rvuMillimeters;
  cmbPageSize.ItemIndex := 1;
end;
{------------------------------------------------------------------------------}
{ Converting mm to pixels }
function MMToPixels(mm, ppi: Integer): Integer;
begin
  Result := Round(mm*5*ppi/127);
end;
{------------------------------------------------------------------------------}
{ Returns distance from edit, converted from mm to pixels }
function TForm1.GetDistance(edit: TEdit; ppi: Integer): Integer;
var mm: Integer;
begin
  mm := StrToIntDef(edit.Text, 0);
  edit.Text := IntToStr(mm);
  Result := MMToPixels(mm, ppi)
end;
{------------------------------------------------------------------------------}
{ Filling PageInfo - calculating size and position of header, footer and main doc }
procedure TForm1.CalcPageLayout(Canvas: TCanvas; var PageInfo: TPageInfo);
var ppix,ppiy: Integer;
begin
  ppix := GetDeviceCaps(Canvas.Handle, LOGPIXELSX);
  ppiy := GetDeviceCaps(Canvas.Handle, LOGPIXELSY);

  PageInfo.PageSize.cx := pdf.Page[0].Width;
  PageInfo.PageSize.cy := pdf.Page[0].Height;

  PageInfo.Header := Rect(-1,-1,-1,-1);
  PageInfo.Footer := Rect(-1,-1,-1,-1);

  PageInfo.Doc.Left  := GetDistance(txtLeft, ppix);
  PageInfo.Doc.Right := PageInfo.PageSize.cx-GetDistance(txtRight, ppix);
  PageInfo.Doc.Top   := GetDistance(txtTop, ppiy);
  PageInfo.Doc.Bottom := PageInfo.PageSize.cy-GetDistance(txtBottom, ppiy);;

  if rvrhHeader.RichView.ItemCount>0 then
  begin
    PageInfo.Header.Left  := PageInfo.Doc.Left;
    PageInfo.Header.Right := PageInfo.Doc.Right;
    rvrhHeader.Init(Canvas, PageInfo.Header.Right-PageInfo.Header.Left);
    rvrhHeader.FormatNextPage($FFFFFFF);
    PageInfo.Header.Top   := GetDistance(txtHeaderY, ppiy);
    PageInfo.Header.Bottom := PageInfo.Header.Top+rvrhHeader.GetLastPageHeight;
    if PageInfo.Header.Bottom>PageInfo.Doc.Top then
      PageInfo.Doc.Top := PageInfo.Header.Bottom;
  end;
  if rvrhFooter.RichView.ItemCount>0 then
  begin
    PageInfo.Footer.Left  := PageInfo.Doc.Left;
    PageInfo.Footer.Right := PageInfo.Doc.Right;
    rvrhFooter.Init(Canvas, PageInfo.Header.Right-PageInfo.Header.Left);
    rvrhFooter.FormatNextPage($FFFFFFF);
    PageInfo.Footer.Bottom := PageInfo.PageSize.cy-GetDistance(txtFooterY, ppiy);
    PageInfo.Footer.Top := PageInfo.Footer.Bottom-rvrhFooter.GetLastPageHeight;
    if PageInfo.Footer.Top<PageInfo.Doc.Bottom then
      PageInfo.Doc.Bottom := PageInfo.Footer.Top;
  end;
  rvrh.Init(Canvas, PageInfo.Doc.Right-PageInfo.Doc.Left);
  while rvrh.FormatNextPage(PageInfo.Doc.Bottom-PageInfo.Doc.Top) do;
end;
{------------------------------------------------------------------------------}
{ Drawing page image on Canvas }
procedure TForm1.DrawPage(Canvas: TCanvas; PageNo: Integer;
  const PageInfo: TPageInfo);
var pt: TPoint;
begin
  SetViewportOrgEx(pdf.Canvas.Handle, PageInfo.Doc.Left, PageInfo.Doc.Top, @pt);
  rvrh.DrawPage(PageNo, Canvas, False,
    PageInfo.Doc.Bottom-PageInfo.Doc.Top);
  if rvrhHeader.RichView.ItemCount>0 then
  begin
    SetViewportOrgEx(pdf.Canvas.Handle, PageInfo.Header.Left, PageInfo.Header.Top, nil);
    rvrhHeader.DrawPage(1, Canvas, False, PageInfo.Header.Bottom-PageInfo.Header.Top);
  end;
  if rvrhFooter.RichView.ItemCount>0 then
  begin
    SetViewportOrgEx(pdf.Canvas.Handle, PageInfo.Footer.Left, PageInfo.Footer.Top, nil);
    rvrhFooter.DrawPage(1, Canvas, False, PageInfo.Footer.Bottom-PageInfo.Footer.Top);
  end;
  SetViewportOrgEx(pdf.Canvas.Handle, pt.X, pt.Y, nil);  
end;
{------------------------------------------------------------------------------}
{ Opening RTF file }
procedure TForm1.btnOpenClick(Sender: TObject);
begin
  rvrh.RichView.Clear;
  rvrhHeader.RichView.Clear;
  rvrhFooter.RichView.Clear;
  rvrh.RichView.DeleteUnusedStyles(True,True,True);
  btnSave.Enabled := False;
  if od.Execute then
  begin
    Screen.Cursor := crHourGlass;
    ShowText('Loading RTF file...', True);
    if rvrh.RichView.LoadRTF(od.FileName) then
      btnSave.Enabled := True
    else
      ShowError('Load Error');
    ShowText('', False);

    txtLeft.Text := IntToStr(Round(rvrh.RichView.DocParameters.LeftMargin));
    txtTop.Text := IntToStr(Round(rvrh.RichView.DocParameters.TopMargin));
    txtRight.Text := IntToStr(Round(rvrh.RichView.DocParameters.RightMargin));
    txtBottom.Text := IntToStr(Round(rvrh.RichView.DocParameters.BottomMargin));
    txtHeaderY.Text := IntToStr(Round(rvrh.RichView.DocParameters.HeaderY));
    txtFooterY.Text := IntToStr(Round(rvrh.RichView.DocParameters.FooterY));
    rgPageOrientation.ItemIndex := ord(rvrh.RichView.DocParameters.Orientation);
    // you can also convert rvrh.RichView.DocParameters.PageWidth and PageHeight
    // to PDF page size
    Screen.Cursor := crDefault;
  end;
end;
{------------------------------------------------------------------------------}
{ Creating and saving PDF }
procedure TForm1.btnSaveClick(Sender: TObject);
var i: Integer;
    PageSize: TPDFPageSize;
    Orientation: TPDFPageOrientation;
const MinDocSize = 50;
begin
  if not sd.Execute then
    exit;

  PageSize := GetPageSize;
  Orientation := GetPageOrientation;

  pdf.Resolution := Screen.PixelsPerInch;
  pdf.FileName := sd.FileName;
  pdf.BeginDoc;

  pdf.Page[0].Size := PageSize;
  pdf.Page[0].Orientation := Orientation;

  Screen.Cursor := crHourGlass;
  ShowText('Formatting...', True);

  CalcPageLayout(pdf.Canvas, FPageInfo);

  if (FPageInfo.Doc.Right-FPageInfo.Doc.Left<MinDocSize) then
  begin
    ShowError('Horizontal margins are too large');
    txtLeft.SetFocus;
    pdf.Abort;
    ShowText('', False);
    Screen.Cursor := crDefault;
    exit;
  end;
  if (FPageInfo.Doc.Bottom-FPageInfo.Doc.Right<MinDocSize) then
  begin
    ShowError('Vertical margins are too large');
    txtTop.SetFocus;
    pdf.Abort;
    ShowText('', False);
    Screen.Cursor := crDefault;
    exit;
  end;

  try
    FCheckpoints := TCollection.Create(TCheckpointLocation);
    FLinksToCheckpoints := TCollection.Create(TLinkToCheckpoint);

    for i := 0 to rvrh.PagesCount-1 do
    begin
      ShowText(Format('Collecting PDF pages (%d of %d)... ', [i, rvrh.PagesCount]), False);
      if i>0 then
      begin
        pdf.NewPage;
        pdf.Page[i].Size        := PageSize;
        pdf.Page[i].Orientation := Orientation;
      end;
      FCurPage := pdf.Page[i];
      FCurPageIndex := i;
      DrawPage(pdf.Canvas, i+1, FPageInfo);
    end;
    ShowText('Writing PDF file...', True);
    SaveLinksToCheckpoints;
    pdf.EndDoc;
  finally
    FreeAndNil(FCheckpoints);
    FreeAndNil(FLinksToCheckpoints);
    Screen.Cursor := crDefault;
    ShowText('', False);
  end;
end;
{------------------------------------------------------------------------------}
{ Returns page orientation }
function TForm1.GetPageOrientation: TPDFPageOrientation;
begin
  Result := TPDFPageOrientation(rgPageOrientation.ItemIndex);
end;
{------------------------------------------------------------------------------}
{ Returns page size }
function TForm1.GetPageSize: TPDFPageSize;
begin
  Result := TPDFPageSize(cmbPageSize.ItemIndex+1);
end;
{------------------------------------------------------------------------------}
{ Displays status text }
procedure TForm1.ShowText(s: String; RedrawForm: Boolean);
begin
  if s='' then
    s := 'Ready';
  lblStatus.Caption := s;
  if RedrawForm then
    Update
  else
    lblStatus.Update;
end;
{------------------------------------------------------------------------------}
{ Displays error message }
procedure TForm1.ShowError(s: String);
begin
  Application.MessageBox(PChar(s), 'Error', MB_OK or MB_ICONSTOP);
end;
{------------------------------------------------------------------------------}
{ Event: rvrv.RichView.OnReadHyperlink.
  Occurs when reading hyperlinks from RTF.
  Saving target in the item tag }
procedure TForm1.rvrhRichViewReadHyperlink(Sender: TCustomRichView;
  const Target, Extras: String; DocFormat: TRVLoadFormat; var StyleNo: Integer;
  var ItemTag: TRVTag; var ItemName: TRVRawByteString);
begin
  ItemTag := Target;
end;
{------------------------------------------------------------------------------}
{ Saving the hyperlink in the RVData's ItemNo-th item in PDF at the location
  specified in R. The location is relative to DocRect.
  If this is a local link (starts from '#'), we add it in FLinksToCheckpoints
  collection. Otherwise, we save the link to PDF as a link to URL. }
procedure TForm1.SetPDFURL(RVData: TCustomRVData; ItemNo: Integer; R, DocRect: TRect);
var Target: String;
begin
  try
    Target := RVData.GetItemTag(ItemNo);
    if Target<>'' then
    begin
      OffsetRect(R, DocRect.Left, DocRect.Top);
      if Target[1]<>'#' then
        FCurPage.SetUrl(R, Target)
      else
        with FLinksToCheckpoints.Add as TLinkToCheckpoint do
        begin
          Area := R;
          PageIndex := FCurPageIndex;
          CheckpointName := Copy(Target, 2, Length(Target)-1);
        end
    end;
  except
  end;
end;
{------------------------------------------------------------------------------}
{ Event: rvrh.OnDrawHyperlink. }
procedure TForm1.rvrhDrawHyperlink(Sender: TRVReportHelper;
  RVData: TCustomRVData; ItemNo: Integer; R: TRect);
begin
  SetPDFURL(RVData, ItemNo, R, FPageInfo.Doc);
end;
{------------------------------------------------------------------------------}
{ Event: rvrhHeader.OnDrawHyperlink. }
procedure TForm1.rvrhHeaderDrawHyperlink(Sender: TRVReportHelper;
  RVData: TCustomRVData; ItemNo: Integer; R: TRect);
begin
  SetPDFURL(RVData, ItemNo, R, FPageInfo.Header);
end;
{------------------------------------------------------------------------------}
{ Event: rvrhFooter.OnDrawHyperlink. }
procedure TForm1.rvrhFooterDrawHyperlink(Sender: TRVReportHelper;
  RVData: TCustomRVData; ItemNo: Integer; R: TRect);
begin
  SetPDFURL(RVData, ItemNo, R, FPageInfo.Footer);
end;
{------------------------------------------------------------------------------}
{ Event: rvrh.OnDrawCheckpoint. Occurs when drawing checkpoint.
  We add the location of this checkpoint to FCheckpoints collection }
procedure TForm1.rvrhDrawCheckpoint(Sender: TRVReportHelper;
  RVData: TCustomRVData; ItemNo, X, Y: Integer);
var CheckpointData: TCheckpointData;
    CPTag: TRVTag;
    CPName: String;
    RE: Boolean;
begin
  CheckpointData := RVData.GetItemCheckpoint(ItemNo);
  RVData.GetCheckpointInfo(CheckpointData, CPTag, CPName, RE);
  with FCheckpoints.Add as TCheckpointLocation do
  begin
    Name := CPName;
    PageIndex := FCurPageIndex;
    Top := Y + FPageInfo.Doc.Top;
  end;
end;
{------------------------------------------------------------------------------}
{ Returns position of checkpoint with the given name }
function TForm1.FindCheckpoint(const CheckpointName: String; var PageIndex,
  Top: Integer): Boolean;
var i: Integer;
begin
  for i := 0 to FCheckpoints.Count-1 do
    if TCheckpointLocation(FCheckpoints.Items[i]).Name=CheckpointName then
    begin
      PageIndex := TCheckpointLocation(FCheckpoints.Items[i]).PageIndex;
      Top := TCheckpointLocation(FCheckpoints.Items[i]).Top;
      Result := True;
      exit;
    end;
  Result := False;
end;
{------------------------------------------------------------------------------}
{ Saves all links to checkpoints to PDF }
procedure TForm1.SaveLinksToCheckpoints;
var i, PageIndex, Top: Integer;
begin
  for i := 0 to FLinksToCheckpoints.Count-1 do
    if FindCheckpoint(TLinkToCheckpoint(FLinksToCheckpoints.Items[i]).CheckpointName,
      PageIndex, Top) then
    begin
      pdf.CurrentPageIndex := TLinkToCheckpoint(FLinksToCheckpoints.Items[i]).PageIndex;
      pdf.Page[TLinkToCheckpoint(FLinksToCheckpoints.Items[i]).PageIndex].SetLinkToPage(
        TLinkToCheckpoint(FLinksToCheckpoints.Items[i]).Area, PageIndex, Top);
    end;
end;
{------------------------------------------------------------------------------}
{ Go to www.llion.net }
procedure TForm1.Image1Click(Sender: TObject);
begin
  ShellExecute(0, 'open', 'http://www.llion.net/?From=RTF2PDF', nil, nil, SW_SHOW);
end;
{------------------------------------------------------------------------------}
{ Go to www.trichview.com }
procedure TForm1.Image2Click(Sender: TObject);
begin
  ShellExecute(0, 'open', 'http://www.trichview.com/?From=RTF2PDF', nil, nil, SW_SHOW);
end;


end.

